@@ -47,6 +47,48 @@ const unsigned char etoa64[] = {
     0060, 0061, 0062, 0063, 0064, 0065, 0066, 0067, 0070, 0071, 0372, 0373, 0374, 0375, 0376, 0377
 };
 
+// parse_hex_digit is a lookup table for finding the numeric values of a hex characters, or otherwise -1
+// it's a bit verbose, but it's fast and it doesn't use ranges
+static const int parse_hex_digit[256] = {
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 0..15
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 16..31
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 32..47
+    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, -1, -1, -1, -1, -1, -1, // 48..63, '0'-'9'
+    -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 64..79, 'A'-'F'
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 80..95
+    -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 96..111, 'a'-'f'
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 112..127
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 128..143
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 144..159
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 160..175
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 176..191
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 192..207
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 208..223
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 224..239
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 // 240..255
+};
+
+// parse_binary_digit is a lookup table for finding the numeric values of binary characters, or otherwise -1
+// it's a bit verbose, but it's fast and it doesn't use ranges
+static const int parse_binary_digit[256] = {
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 0..15
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 16..31
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 32..47
+    0, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 48..63, '0'-'1'
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 64..79
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 80..95
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 96..111
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 112..127
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 128..143
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 144..159
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 160..175
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 176..191
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 192..207
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 208..223
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, // 224..239
+    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 // 240..255
+};
+
 void exit_with_usage(const char* program_name, const char* version)
 {
     fprintf(stderr,
@@ -56,7 +98,7 @@ void exit_with_usage(const char* program_name, const char* version)
         "       %s -r [-s [-]offset] [-c cols] [-ps] [infile [outfile]]\n"
         "Options:\n"
         "    -a          toggle autoskip: A single '*' replaces nul-lines. Default off.\n"
-        "    -b          binary digit dump (incompatible with -ps,-i). Default hex.\n"
+        "    -b          binary digit dump (incompatible with -ps). Default hex.\n"
         "    -C          capitalize variable names in C include file style (-i).\n"
         "    -c cols     format <cols> octets per line. Default 16 (-i: 12, -ps: 30).\n"
         "    -E          show characters in EBCDIC. Default ASCII.\n"
@@ -134,21 +176,6 @@ void fclose_or_die(const char* program_name)
     }
 }
 
-// parse_hex_digits returns the decimal value if c is a hex digit, or otherwise -1.
-inline int parse_hex_digit(const int ch)
-{
-    return (ch >= '0' && ch <= '9') ? ch - '0'
-        : (ch >= 'a' && ch <= 'f')  ? ch - 'a' + 10
-        : (ch >= 'A' && ch <= 'F')  ? ch - 'A' + 10
-                                    : -1;
-}
-
-// parse_bin_digit returns the decimal value if c is a binary digit, or otherwise -1.
-inline int parse_bin_digit(const int ch)
-{
-    return (ch < '0' || ch > '1') ? -1 : ch - '0';
-}
-
 // skip_to_eol_or_die will ignore text from the input file, until EOL or EOF
 static inline void skip_to_eol_or_die(int* ch, const char* program_name)
 {
@@ -184,8 +211,11 @@ int decode_hex_stream_postscript(const long base_off, const char* program_name)
     int c = 0, n1 = -1, n2 = 0, n3 = 0, tmp = -1;
     long have_off = 0, want_off = 0;
     rewind(input_file);
-    while (((c = getc(input_file)) != EOF) && c != ' ' && c != '\n' && c != '\t' && c != '\r') {
-        if ((tmp = parse_hex_digit(c)) == -1 && ignore) {
+    while (((c = getc(input_file)) != EOF)) {
+        if (c == ' ' || c == '\n' || c == '\t' || c == '\r') {
+            continue;
+        }
+        if ((tmp = parse_hex_digit[c]) == -1 && ignore) {
             continue;
         }
         n3 = n2;
@@ -218,7 +248,7 @@ int decode_hex_stream_normal(const int cols, const long base_off, const char* pr
     long have_off = 0, want_off = 0;
     rewind(input_file);
     while (((c = getc(input_file)) != EOF) && c != '\r') {
-        if ((tmp = parse_hex_digit(c)) == -1 && ignore) {
+        if ((tmp = parse_hex_digit[c]) == -1 && ignore) {
             continue;
         }
         n3 = n2;
@@ -269,10 +299,10 @@ int decode_hex_stream_bits(const int cols, const char* program_name)
     long want_off = 0;
     rewind(input_file);
     while (((c = getc(input_file)) != EOF) && c != '\r') {
-        if ((n1 = parse_hex_digit(c)) == -1 && ignore) {
+        if ((n1 = parse_hex_digit[c]) == -1 && ignore) {
             continue;
         }
-        bit = parse_bin_digit(c);
+        bit = parse_binary_digit[c];
         if (bit != -1) {
             bit_buffer = ((bit_buffer << 1) | bit);
             bit_count++;
@@ -317,7 +347,6 @@ int decode_hex_stream_bits(const int cols, const char* program_name)
 static inline void print_or_suppress_zero_line(const char* buffer, char* z, const int nz, const char* program_name)
 {
     static int zero_seen = 0;
-
     if (nz) {
         if (nz < 0) {
             zero_seen--;
@@ -397,7 +426,7 @@ inline enum ColorDigit ascii_char_color(const unsigned char e)
     return COLOR_RED;
 }
 
-int hex_postscript(const bool colsgiven, int cols, int octspergrp, const bool revert, int e, const long length, const int negseek, const long seekoff, const char* hex_digits, const char* program_name)
+int hex_postscript(const bool colsgiven, int cols, const bool revert, int e, const long length, const int negseek, const long seekoff, const char* hex_digits, const char* program_name)
 {
     if (!colsgiven) {
         cols = 30;
@@ -407,9 +436,6 @@ int hex_postscript(const bool colsgiven, int cols, int octspergrp, const bool re
     if (revert) {
         return decode_hex_stream_postscript(negseek ? -seekoff : seekoff, program_name);
     }
-    if (octspergrp < 1 || octspergrp > cols) {
-        octspergrp = cols;
-    }
     int n = 0, p = cols;
     getc_or_die(&e, program_name);
     while ((length < 0 || n < length) && e != EOF) {
@@ -428,7 +454,7 @@ int hex_postscript(const bool colsgiven, int cols, int octspergrp, const bool re
     return 0;
 }
 
-int hex_cinclude(const bool colsgiven, int cols, int octspergrp, const bool revert, int e, int c, const bool capitalize, const char* varname, const char* argv1, const bool uppercase_hex, const long length, const char* program_name)
+int hex_cinclude(const bool colsgiven, int cols, const bool revert, int e, int c, const bool capitalize, const char* varname, const char* argv1, const bool uppercase_hex, const long length, const char* program_name)
 {
     int p = 0;
     if (!colsgiven || !cols) {
@@ -439,9 +465,6 @@ int hex_cinclude(const bool colsgiven, int cols, int octspergrp, const bool reve
     if (revert) {
         exit_with_error(-1, "Sorry, cannot revert this type of hexdump", program_name);
     }
-    if (octspergrp < 1 || octspergrp > cols) {
-        octspergrp = cols;
-    }
     if (!varname && input_file != stdin) {
         varname = argv1; // argv[1]
     }
@@ -489,6 +512,71 @@ int hex_cinclude(const bool colsgiven, int cols, int octspergrp, const bool reve
     return 0;
 }
 
+int hex_cinclude_bits(const bool colsgiven, int cols, const bool revert, int e, int c, const bool capitalize, const char* varname, const char* argv1, const long length, const char* program_name)
+{
+    int p = 0;
+    if (!colsgiven || !cols) {
+        cols = 6;
+    } else if (cols < 1) {
+        exit_with_col_error(program_name);
+    }
+    if (revert) {
+        exit_with_error(-1, "Sorry, cannot revert this type of hexdump", program_name);
+    }
+    if (!varname && input_file != stdin) {
+        varname = argv1; // argv[1]
+    }
+    if (varname) {
+        if (fprintf(output_file, "unsigned char %s", isdigit((unsigned char)varname[0]) ? "__" : "") < 0) {
+            exit_with_error(3, NULL, program_name);
+        }
+        for (e = 0; (c = varname[e]); e++) {
+            if (!capitalize) {
+                putc_or_die(isalnum((unsigned char)c) ? c : '_', program_name);
+            } else {
+                putc_or_die(isalnum((unsigned char)c) ? toupper((unsigned char)(c)) : '_', program_name);
+            }
+        }
+        fputs_or_die("[] = {\n", program_name);
+    }
+    getc_or_die(&c, program_name);
+    while ((length < 0 || p < length) && c != EOF) {
+        if (p == 0) {
+            fputs_or_die("  ", program_name);
+        } else if (p % cols == 0) {
+            fputs_or_die(",\n  ", program_name);
+        } else {
+            fputs_or_die(", ", program_name);
+        }
+        fputs_or_die("0b", program_name);
+        for (int bit = 7; bit >= 0; bit--) {
+            putc_or_die(((c >> bit) & 1) + '0', program_name);
+        }
+        p++;
+        getc_or_die(&c, program_name);
+    }
+    if (p) {
+        putc_or_die('\n', program_name);
+    }
+    if (varname) {
+        fputs_or_die("};\n", program_name);
+        if (fprintf(output_file, "unsigned int %s", isdigit((unsigned char)varname[0]) ? "__" : "") < 0) {
+            exit_with_error(3, NULL, program_name);
+        }
+        for (e = 0; (c = varname[e]); e++) {
+            if (!capitalize) {
+                putc_or_die(isalnum((unsigned char)c) ? c : '_', program_name);
+            } else {
+                putc_or_die(isalnum((unsigned char)c) ? toupper((unsigned char)(c)) : '_', program_name);
+            }
+        }
+        if (fprintf(output_file, "_%s = %d;\n", capitalize ? "LEN" : "len", p) < 0) {
+            exit_with_error(3, NULL, program_name);
+        }
+    }
+    return 0;
+}
+
 int hex_bits(char* buffer, char* z, const bool colsgiven, int cols, int octspergrp, const bool revert, int c, int e, const long length, const char* decimal_format_string, const long seekoff, const unsigned long displayoff, const bool color, const bool ascii, const bool autoskip, const char* program_name)
 {
     int grplen = 0, n = 0, nonzero = 0, p = 0, addrlen = 9;
@@ -515,8 +603,9 @@ int hex_bits(char* buffer, char* z, const bool colsgiven, int cols, int octsperg
         }
         c = addrlen + 1 + (grplen * p) / octspergrp;
         for (int i = 7; i >= 0; i--) {
-            buffer[c++] = (e & (1 << i)) ? '1' : '0';
+            buffer[c++] = ((e >> i) & 1) + '0';
         }
+
         if (color) {
             c = (grplen * cols - 1) / octspergrp + addrlen + 3 + p * 12;
             nonzero += e ? 1 : 0;
@@ -724,7 +813,7 @@ int hex_littleendian(char* buffer, char* z, const bool colsgiven, int cols, int
             if (!ascii) { // EBCDIC
                 e = (e < 64) ? '.' : etoa64[e - 64];
             }
-            c += addrlen + 3 + p;
+            c += addrlen + 2 + p;
             buffer[c++] = (e < ' ' || e >= 127) ? '.' : e;
         }
         n++;
@@ -776,13 +865,14 @@ const char* base_name(const char* path)
 
 int main(int argc, char* argv[])
 {
-    const char* version = "tinyxxd 1.3.3";
+    const char* version = "tinyxxd 1.3.7";
     enum HexType {
         HEX_NORMAL,
         HEX_BITS,
         HEX_CINCLUDE,
         HEX_LITTLEENDIAN,
-        HEX_POSTSCRIPT
+        HEX_POSTSCRIPT,
+        HEX_BITS_AND_CINCLUDE
     };
     enum HexType hextype = HEX_NORMAL;
     const char* no_color = getenv("NO_COLOR"); // Respect the NO_COLOR environment variable
@@ -801,7 +891,11 @@ int main(int argc, char* argv[])
         if (!strncmp(pp, "-a", 2)) {
             autoskip = !autoskip;
         } else if (!strncmp(pp, "-b", 2)) {
-            hextype = HEX_BITS;
+            if (hextype == HEX_CINCLUDE) {
+                hextype = HEX_BITS_AND_CINCLUDE;
+            } else {
+                hextype = HEX_BITS;
+            }
         } else if (!strncmp(pp, "-e", 2)) {
             hextype = HEX_LITTLEENDIAN;
         } else if (!strncmp(pp, "-u", 2)) {
@@ -809,7 +903,11 @@ int main(int argc, char* argv[])
         } else if (!strncmp(pp, "-p", 2)) {
             hextype = HEX_POSTSCRIPT;
         } else if (!strncmp(pp, "-i", 2)) {
-            hextype = HEX_CINCLUDE;
+            if (hextype == HEX_BITS) {
+                hextype = HEX_BITS_AND_CINCLUDE;
+            } else {
+                hextype = HEX_CINCLUDE;
+            }
         } else if (!strncmp(pp, "-C", 2)) {
             capitalize = true;
         } else if (!strncmp(pp, "-d", 2)) {
@@ -998,10 +1096,12 @@ int main(int argc, char* argv[])
     case HEX_BITS:
         return hex_bits(buffer, z, colsgiven, cols, octspergrp, revert, ch, e, length, decimal_format_string, seekoff, displayoff, color, ascii, autoskip, program_name);
     case HEX_CINCLUDE:
-        return hex_cinclude(colsgiven, cols, octspergrp, revert, e, ch, capitalize, varname, argv[1], uppercase_hex, length, program_name);
+        return hex_cinclude(colsgiven, cols, revert, e, ch, capitalize, varname, argv[1], uppercase_hex, length, program_name);
     case HEX_LITTLEENDIAN:
         return hex_littleendian(buffer, z, colsgiven, cols, octspergrp, revert, seekoff, color, e, length, decimal_format_string, displayoff, ch, ascii, hex_digits, autoskip, program_name);
     case HEX_POSTSCRIPT:
-        return hex_postscript(colsgiven, cols, octspergrp, revert, e, length, negseek, seekoff, hex_digits, program_name);
+        return hex_postscript(colsgiven, cols, revert, e, length, negseek, seekoff, hex_digits, program_name);
+    case HEX_BITS_AND_CINCLUDE:
+        return hex_cinclude_bits(colsgiven, cols, revert, e, ch, capitalize, varname, argv[1], length, program_name);
     }
 }
